//=========================================================================
//Copyright LizardOfOz.
//This is a VERY early version intended to prototype maps.
//Expect things to change over time.
//
//Check the following links for updates:
//   https://tf2maps.net/members/lizardofoz.39426/#resources
//   https://twitter.com/LizardOfOz_
//   https://youtube.com/LizardOfOz
//
//Credits:
//  LizardOfOz - the original VSH Plugin. VSH Vscript recreation.
//  Maxxy, Velly, JPRAS - Saxton Hale's model imitating the Jungle Inferno short.
//  MegapiemanPHD - Hale's voice lines.
//  Yakibomb - give_tf_weapon script bundle (used for Hale's first-person hands model).
//=========================================================================

isRoundSetup <- true
isRoundOver <- false;
starterClass <- {}
isFeedbackRound <- false;

function StartNewRound()
{
    DoEntFire("game_text_victory_screen", "AddOutput", "message ", 0, null, null);

    local roundTimer = Entities.FindByClassname(null, "team_round_timer");
    if (roundTimer != null)
        roundTimer.Kill();
    roundTimer = SpawnEntityFromTable("team_round_timer", {
        targetname = "team_round_timer",
        auto_countdown = 1,
        max_length = 0,
        reset_time = 1,
        setup_length = setupLength,
        show_in_hud = 1,
        show_time_remaining = 1,
        start_paused = 0,
        timer_length = 240,
        StartDisabled = 0,
        "OnSetupFinished#1": "!self,RunScriptFile,vssaxtonhale/external_calls/finish_round_setup.nut,0,-1",
        "OnSetupFinished#2": "vsh_setup_end_relay,Trigger,,0,-1",
        "OnFinished": "!self,RunScriptFile,vssaxtonhale/external_calls/unlock_control_point.nut,0,-1"
    });
    EntFireByHandle(roundTimer, "Resume", "", 0, null, null);

    DoEntFire("vsh_round_start_relay", "Trigger", "", 0, null, null);
    DoEntFire("vsh_setup_start_relay", "Trigger", "", 0, null, null);

    SetRoundSetup(true);
    isRoundOver = false;

    FireListeners("start", null);
}

function StartFirstNewRound(void)
{
    StartNewRound();
}
AddListener("map_start", StartFirstNewRound, 1);

function OnNewRoundStarted(void)
{
    if (isFeedbackRound || IsInWaitingForPlayers() || GetTotalPlayerCount() < 2)
        return;
    ResetRoundDamage();

    isRoundSetup = true;

    ProgressHaleQueue();

    local forceRespawn = Entities.FindByClassname(null, "game_forcerespawn");
    EntFireByHandle(forceRespawn, "ForceRespawn", "", 0, null, null);

    foreach (player in GetAllValidMercs())
    {
        player.ForceChangeTeam(TF_TEAM.MERCS, true);
        player.ForceRegenerateAndRespawn();
        player.KeyValueFromString("targetname", "");
        TrySpawnAsMerc(player);
    }

    haleConversionAttempts = 0;
    TrySpawnAsHale(GetHale());
    RunWithDelay("PlaySoundForAll(`VSH.Start`)", null, 3);
}
AddListener("start", OnNewRoundStarted, 1);

function SetRoundSetup(isSetup)
{
    isRoundSetup = isSetup;
    TrySpawnAsHale(GetHale());

    if (isSetup)
        return;

    foreach (player in GetAllValidMercs())
        starterClass[player] <- player.GetPlayerClass();

    local respawnRoom = null;
    while (respawnRoom = Entities.FindByClassname(respawnRoom, "func_respawnroom"))
        respawnRoom.SetAbsOrigin(respawnRoom.GetOrigin() + Vector(0,0,-10000));
}

function ProgressHaleQueue()
{
    local gamerules = Entities.FindByClassname(null, "tf_gamerules");
    if (NetProps.GetPropInt(gamerules, "m_nHudType") != 2) //todo move to a better place
        NetProps.SetPropInt(gamerules, "m_nHudType", 2);
    gamerules.ValidateScriptScope();
    local globalVars = gamerules.GetScriptScope();

    if ("nextHale" in globalVars && IsValidPlayer(globalVars.nextHale))
    {
        SetHale(globalVars.nextHale);
        globalVars.lastHale <- GetHale();
    }
    else
    {
        local newHale = RollNewHale("lastHale" in globalVars ? globalVars.lastHale : null);
        globalVars.lastHale <- newHale;
        SetHale(newHale);
    }
    globalVars.nextHale <- null;
}

function RollNewHale(ignoredPlayer)
{
    local players = [];
    foreach (player in GetAllValidPlayers())
        if (player != ignoredPlayer)
            players.push(player);

    if (players.len() > 0)
        return players[RandomInt(0, players.len() - 1)];

    local theOnlyPlayer = Entities.FindByClassname(null, "player");
    return (theOnlyPlayer != null && !theOnlyPlayer.IsFakeClient()) ? theOnlyPlayer : null ;
}

function HasRoundEnded(void)
{
    if (IsInWaitingForPlayers())
        return;
    if (GetAliveMercCount() == 0 && GetTotalPlayerCount() >= 2)
        EndRound(TF_TEAM.HALE);
    local hale = GetHale();
    if (!IsRoundSetup() && (IsNotValidPlayer(hale) || IsPlayerDead(hale)))
        EndRound(TF_TEAM.MERCS);
}
AddListener("tick_always", HasRoundEnded, 1);

function EndRound(winner)
{
    if (isRoundOver)
        return;
    isRoundOver = true;
    PlaySoundForAll(winner == TF_TEAM.HALE ? "VSH.HaleWins" : "VSH.HaleLoses");
    local gamerules = Entities.FindByClassname(null, "tf_gamerules");
    EntFireByHandle(gamerules, winner == TF_TEAM.HALE ? "AddBlueTeamScore" : "AddRedTeamScore", "1", 3, null, null);
    RunWithDelay("PlaySoundForAll(\"Hud.EndRoundScored\")", null, 3);
    DisplayDamageBoard(winner);
    RunWithDelay("DisplayDamageBoard("+winner+")", null, 3.1);

    local startTime = NetProps.GetPropFloat(gamerules, "m_flMapResetTime");
    local roundsPlayed = NetProps.GetPropFloat(gamerules, "m_nRoundsPlayed");
    local timeLimit = Convars.GetInt("mp_timelimit");
    local maxRounds = Convars.GetInt("mp_maxrounds");
    local timePlayed = Time() - startTime;
    local timeLeft = timeLimit * 60.0 - timePlayed;

    //Prioritize `mp_timelimit`. If set to 0, check `mp_maxrounds` and play 4 times as many rounds.
    //Why 4 times as many? Because VSH rounds last only 3 minutes on average, 4 times less than a typical round of payload.
    //todo roun limit doesn't work
    if ((timeLimit > 0 && timeLeft <= 0) || (timeLimit <= 0 && maxRounds > 0 && roundsPlayed >= maxRounds * 4))
    {
        if (IsPlayingOnTF2MapsNet())
        {
            local roundWin = SpawnEntityFromTable("game_round_win",
            {
                win_reason = "0",
                force_map_reset = "1", //not having
                TeamNum = winner,      //these 3 lines
                switch_teams = "0"     //causes the crash when trying to fire game_round_win
            });
            EntFireByHandle(roundWin, "RoundWin", "", 0, null, null);
        }
        else
        {
            local gameEnd = SpawnEntityFromTable("game_end", {});
            EntFireByHandle(gameEnd, "EndGame", "", 10, null, null);
        }
    }
    else
    {
        local controlPoint = Entities.FindByClassname(null, "team_control_point");
        EntFireByHandle(controlPoint, "SetOwner", "0", 9.8, controlPoint, controlPoint);
        EntFireByHandle(controlPoint, "SetLocked", "1", 9.8, null, null);
        local respawnRoom = null;
        while (respawnRoom = Entities.FindByClassname(respawnRoom, "func_respawnroom"))
            respawnRoom.SetAbsOrigin(respawnRoom.GetOrigin() + Vector(0,0,10000));
        RunWithDelay("StartNewRound()", null, 10);
    }
}

function DisplayDamageBoard(winner)
{
    local hale = GetHale();
    local haleHP = IsValidPlayer(hale) ? IsPlayerDead(hale) ? 0 : hale.GetHealth() : "N/A";
    local text1 = "Damage dealt this round∶";
    local text2 = "Most damage dealt by∶";
    local i = 0;
    foreach (entry in GetDamageBoardSorted())
    {
        local player = entry[0];
        local damage = entry[1];
        if (!IsValidPlayer(player))
            continue;
        local name = NetProps.GetPropString(player, "m_szNetname");
        local clazz = TF_CLASS_NAMES[player in starterClass ? starterClass[player] : 0];
        text1 += "\n"+damage+" - "+name+" ("+clazz+")";
        if (i++ < 3)
            text2 += "\n"+damage+" - "+name+" ("+clazz+")";
    }
    if (i == 0)
        text1 = text2 = (winner == TF_TEAM.MERCS ? "Hale bailed out" : "You're all losers!");
    Say(null, "Hale's health was: " + haleHP + "/" + haleStartHealth+"\n"+text1, false);

    text2+="\n\nScore∶\nHale  |  Mercs\n";

    local teamEntity = null;
    while (teamEntity = Entities.FindByClassname(teamEntity, "tf_team"))
    {
        if (NetProps.GetPropInt(teamEntity, "m_iTeamNum") == TF_TEAM.HALE)
            text2+=NetProps.GetPropInt(teamEntity, "m_iScore");
    }
    teamEntity = null;
    while (teamEntity = Entities.FindByClassname(teamEntity, "tf_team"))
    {
        if (NetProps.GetPropInt(teamEntity, "m_iTeamNum") == TF_TEAM.MERCS)
            text2+="          " + NetProps.GetPropInt(teamEntity, "m_iScore");
    }
    DoEntFire("game_text_victory_screen", "AddOutput", "message "+text2, 0, null, null);
}
